#include "wl_def.h"

/* mapdraw.c */

/* This code is originally from the author of XWolf. */

/* displays Wolfenstein 3D maps */

#include <fcntl.h>
#include <png.h>

/* ---------------- Map data ----------------- */

#if 0
/* from wl_def.h */
#define AREATILE 107	 // first of NUMAREAS floor tiles
#define NUMAREAS 37
#define ELEVATORTILE	21
#define	ALTELEVATORTILE	107

/* from id_ca.h */
#define NUMMAPS		60
#define MAPPLANES	2

char extension[] = "SDM";
#endif

int lastmap = 0;

#define CARMACIZED

/* the structure of the MAPHEAD.xxx file */
typedef struct
{
	unsigned short	RLEWtag;
	long		headeroffsets[100];
	byte		tileinfo[1];
} __attribute__ ((packed)) mapfiletype ;

mapfiletype *maphead;

#if 0

/* the map header in GAMEMAPS.xxx */
typedef	struct
{
	long		planestart[3];
	unsigned short	planelength[3];
	unsigned short	width,height;
	char		name[16];
} maptype;

maptype	*mapheaderseg[NUMMAPS];

unsigned short	*mapsegs[MAPPLANES];
#endif

/* ---------------- image data ---------------- */

#define TILE_SIZE 8

png_structp png_ptr;
png_infop info_ptr;
png_bytep row_pointer;

FILE *fp;


/* ---------------- option data --------------- */

#if 0
struct option longopts [] =
{
    {"help", 0, NULL, 0},
    {"map",  1, NULL, 1},
    {"mapdraw",  0, NULL, 2},
    {"head", 0, NULL, 3},
    {"png",  0, NULL, 4},
    {"text", 0, NULL, 5},
    {0, 0, 0, 0}
};
#endif

int show_head = 0;
int do_png = 1;
int do_text = 0;

/* ---------------- misc routines --------------- */

static void
fail (char *mess)
{
    puts (mess);
    exit (-1);
}

/*
static int
filelength (int handle)
{
    struct stat statbuf;
    if (fstat (handle, &statbuf) < 0) {
	perror ("filelength: ");
	exit (-1);
    }
    return statbuf.st_size;
}
*/

static void
getptr (void **pp, int size)
{
    void *p;
    p = malloc (size);
    if (p == NULL) {
	printf ("malloc failed\n");
	exit (-1);
    }
    *pp = p;
}

/* ------------- map routines --------------- */

static void
read_maphead (void)
{
    int n;
    int handle;
    long length;
    char fname[MAX_PATH+1];


    /* read the header file */

#ifdef __unix__
    snprintf(fname, MAX_PATH, "%s/%s/%s/%s%s", getenv("HOME"), CONFIG_USER_DIR, mod_extension, mheadname, mod_extension);
#else
    strcpy (fname, mheadname);
    strcat (fname, extension);
#endif

    handle = open (fname, O_RDONLY);
    if (handle == -1) {
#ifdef __unix__
	sprintf(fname, "%s/%s%s", wolfdatadir, mheadname, mod_extension);
	handle = OpenRead(fname);
		if (handle == -1) {
			CA_CannotOpen(fname);
		}
#else
	CA_CannotOpen(fname);
#endif
    }
    length = filelength (handle);
    maphead = (mapfiletype *) malloc (length);
    if (maphead == NULL) {
	printf ("malloc failed for maphead\n");
	exit (-1);
    }
    if (show_head)
	printf ("reading %ld bytes from %s\n", length, fname);
    n = read (handle, maphead, length);
    if (n != length) {
	printf ("can't read from %s\n", fname);
	exit (-1);
    }
    close (handle);
}

int maphandle;

static void
read_mapdata (void)
{
    int i;
    char fname[13];
    long pos;

#ifdef CARMACIZED
#ifdef __unix__
    snprintf(fname, MAX_PATH, "%s/%s/%s/%s%s", getenv("HOME"), CONFIG_USER_DIR, mod_extension, gmapsname, mod_extension);
#else
    strcpy (fname, gmapsname);
    strcat (fname, extension);
#endif

    maphandle = open(fname, O_RDONLY);
    if (maphandle == -1) {
#ifdef __unix__
	sprintf(fname, "%s/%s%s", wolfdatadir, gmapsname, mod_extension);
	maphandle = OpenRead(fname);
	if (maphandle == -1) {
		CA_CannotOpen(fname);
	}
#else
	CA_CannotOpen(fname);
#endif
    }
#else
#ifdef __unix__
    snprintf(fname, MAX_PATH, "%s/%s/%s/%s%s", getenv("HOME"), CONFIG_USER_DIR, mod_extension, mfilename, mod_extension);
#else
    strcpy(fname,mfilename);
    strcat(fname,extension);
#endif

    if ((maphandle = open(fname, O_RDONLY)) == -1)
	CA_CannotOpen(fname);
#endif

//
// load all map header
//
    for (i=0; i<NUMMAPS; i++)
    {
	pos = maphead->headeroffsets[i];
	if (pos<=0)
	    break;

	lastmap = i;

	if (show_head) {
	    printf ("map %d\n", i);
	    printf ("  header at %ld\n", pos);
	}

	mapheaderseg[i] = malloc (sizeof(maptype));
	lseek (maphandle, pos, SEEK_SET);
	read (maphandle, mapheaderseg[i], sizeof(maptype));

	if (show_head) {
	    printf ("    planestart[0]  %5ld\n", mapheaderseg[i]->planestart[0]);
	    printf ("    planestart[1]  %5ld\n", mapheaderseg[i]->planestart[1]);
	    printf ("    planelength[0] %5d\n", mapheaderseg[i]->planelength[0]);
	    printf ("    planelength[1] %5d\n", mapheaderseg[i]->planelength[1]);
	    printf ("    width          %5d\n", mapheaderseg[i]->width);
	    printf ("    height         %5d\n", mapheaderseg[i]->height);
	    printf ("    name '%s'\n", mapheaderseg[i]->name);
	}
    }
//
// allocate space for 3 64*64 planes
//
	for (i=0;i<MAPPLANES;i++)
	{
		getptr ((memptr *)&mapsegs[i],64*64*2);
	}
}


static void
read_map (int mapnum)
{
	long	pos,compressed;
	int		plane;
	memptr	*dest,bigbufferseg;
	unsigned	size;
	unsigned short *source;
#ifdef CARMACIZED
	memptr buffer2seg;
	long	expanded;
#endif

/*  	printf ("map %d '%s'\n", mapnum, mapheaderseg[mapnum]->name); */

//
// load the planes into the allready allocated buffers
//
	size = 64*64*2;

	for (plane = 0; plane<MAPPLANES; plane++)
	{

	    pos = mapheaderseg[mapnum]->planestart[plane];
	    compressed = mapheaderseg[mapnum]->planelength[plane];

		dest = (memptr *)&mapsegs[plane];

		lseek(maphandle,pos,SEEK_SET);
		getptr (&bigbufferseg, compressed);
		source = bigbufferseg;

		read (maphandle, (byte *)source, compressed);
#ifdef CARMACIZED
		//
		// unhuffman, then unRLEW
		// The huffman'd chunk has a two byte expanded length first
		// The resulting RLEW chunk also does, even though it's not really
		// needed
		//
		expanded = *source;
		source++;

		getptr (&buffer2seg, expanded);

		CAL_CarmackExpand ((byte *)source, (word *)buffer2seg,expanded);
		CA_RLEWexpand (((unsigned short *)buffer2seg)+1,*dest,size,
			       maphead->RLEWtag);
		free (buffer2seg);

#else
		//
		// unRLEW, skipping expanded length
		//
		CA_RLEWexpand (source+1, *dest,size,
		((mapfiletype *)tinf)->RLEWtag);
#endif

		free (bigbufferseg);
	}
}

/* ------------ image routines ------------ */

static void
setup_png_file (char *filename)
{
    fp = fopen (filename, "w");
    if (!fp)
	fail ("fopen");

    png_ptr = png_create_write_struct
	(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr)
       fail ("png_ptr");

    info_ptr = png_create_info_struct (png_ptr);
    if (!info_ptr) {
       png_destroy_write_struct (&png_ptr, (png_infopp)NULL);
       fail ("info_ptr");
    }

    if (setjmp (png_ptr->jmpbuf)) {    
        png_destroy_write_struct (&png_ptr, &info_ptr);
        fclose (fp);
        fail ("setjmp");
    }

    png_init_io (png_ptr, fp);

    png_set_IHDR (png_ptr, info_ptr,
		  64*TILE_SIZE+1,               /* width */
		  64*TILE_SIZE+1,               /* height */
		  8,                            /* bit_depth*/
		  PNG_COLOR_TYPE_RGB,           /* color_type */
		  PNG_INTERLACE_NONE,           /* interlace_type */
		  PNG_COMPRESSION_TYPE_DEFAULT, /* compression_type*/
		  PNG_FILTER_TYPE_DEFAULT);     /* filter_type*/

    png_write_info (png_ptr, info_ptr);

}


char *rowptrs[64*TILE_SIZE+1];

static void
allocate_image (void)
{
    int row;

    for (row=0; row<64*TILE_SIZE+1; row++) {
	rowptrs [row] = malloc ((64*TILE_SIZE+1)*3);
	if (rowptrs [row] == NULL)
	    fail ("malloc");
    }
}


static void
write_png_image (void)
{
    int row;

    for (row=0; row<64*TILE_SIZE+1; row++) {
	png_write_row (png_ptr, rowptrs[row]);
    }

    png_write_end (png_ptr, info_ptr);
}


static void
fill_square (int x, int y, int red, int green, int blue)
{
    int r, c;
    char *row;

    for (r=0; r<TILE_SIZE; r++) {
	row = rowptrs [y*TILE_SIZE+r];
	for (c=0; c<TILE_SIZE; c++) {
	    row [(x*TILE_SIZE+c)*3+0] = red;
	    row [(x*TILE_SIZE+c)*3+1] = green;
	    row [(x*TILE_SIZE+c)*3+2] = blue;
	}
    }
}


static void
draw_blob (int x, int y, int red, int green, int blue)
{
    int r, c;
    char *row;

    for (r=3; r<TILE_SIZE-2; r++) {
	row = rowptrs [y*TILE_SIZE+r];
	for (c=3; c<TILE_SIZE-2; c++) {
	    row [(x*TILE_SIZE+c)*3+0] = red;
	    row [(x*TILE_SIZE+c)*3+1] = green;
	    row [(x*TILE_SIZE+c)*3+2] = blue;
	}
    }

}


static void
draw_grid (void)
{
    int x, y;
    int r, c;
    char *row;

    /* horizontal lines */
    for (y=0; y<65; y++) {
	row = rowptrs [y*TILE_SIZE];
	for (x=0; x<64; x++) {
	    for (c=0; c<TILE_SIZE; c++) {
		row [(x*TILE_SIZE+c)*3+0] = 70;
		row [(x*TILE_SIZE+c)*3+1] = 70;
		row [(x*TILE_SIZE+c)*3+2] = 70;
	    }
	}
    }

    /* vertical lines */
    for (x=0; x<65; x++) {
	for (y=0; y<64; y++) {
	    for (r=0; r<TILE_SIZE; r++) {
		row = rowptrs [y*TILE_SIZE+r];
		row [(x*TILE_SIZE)*3+0] = 70;
		row [(x*TILE_SIZE)*3+1] = 70;
		row [(x*TILE_SIZE)*3+2] = 70;
	    }
	}
    }
}


static void
draw_walls (void)
{
    short *p0, *p1;
    int x, y;
    int tile, object;

    p0 = mapsegs[0];
    p1 = mapsegs[1];

    for (y=0; y<64; y++) {

	for (x=0; x<64; x++) {

	    tile = p0[y*64+x];
	    object = p1[y*64+x];

	    if (tile<AREATILE) {
		/* walls, doors */
		if (tile >= 90 && tile <= 101) {
		    /* doors */
		    fill_square (x, y, 
				 (tile&0x3F)*3,        /* red */
				 50+(tile&0x3F)*2,     /* green */
				 200);                 /* blue */
		}
		else {
		    /* walls */
		    if (tile == ELEVATORTILE) {
			fill_square (x, y, 255, 255, 255);
		    }
		    else {
			fill_square (x, y, 
				     (tile&0x3F)*3,        /* red */
				     200,                  /* green */
				     50+(tile&0x3F)*2);    /* blue */
		    }
		}
	    }
	    else if (tile<AREATILE+NUMAREAS) {
		/* floor */
		fill_square (x, y, 128, 128, 128);
	    }
	    else {
		/* error */
		fill_square (x, y, 0, 0, 0);
	    }

	    switch (object) {
	      case 0:
		  break;
	      case 19:
	      case 20:
	      case 21:
	      case 22:
		  /* start position */
		  draw_blob (x, y, 255, 0, 0);     /* red */
		  break;
  	      case 99:
		  /* victory tile */
		  draw_blob (x, y, 255, 255, 255); /* white */
		  break;
	      default:
		  draw_blob (x, y, 255, 215, 0);   /* gold */
		  break;
	    }
	}
    }
}


/* ---------------- text output --------------- */

static void
print_walls (void)
{
    short *p;
    int x, y;
    int tile;

    p = mapsegs[0];

    for (y=0; y<64; y++) {
	printf ("%2d: ", y);
	for (x=0; x<64; x++) {
	    tile = p[y*64+x];
	    if (tile<AREATILE)
		printf ("%02x", tile);
	    else if (tile<AREATILE+NUMAREAS)
		printf ("  ");            /* floor area */
	    else
		printf ("??");
	}
	printf ("\n");
    }
}


/* ----------------- main --------------- */

static void
usage (void)
{
    puts ("Usage: mapdraw [OPTION]...\n"
	  "Show information about Wolfenstein 3D maps\n"
	  "\n"
	  "  --help      Show this help\n"
	  "  --which M     dump map number M (default: all maps)\n"
	  "  --head      show map header information\n"
	  "  --png       draw the map as PNG image file to map_XXX_M.png (default)\n"
	  "  --text      draw the map as hex ascii to stdout\n"
	  "\n");
    exit (-1);    
}


void
mapdraw_main ()
{
    int map = -1;
    int i, iters;
    char outfile [MAX_PATH+1];

	if (MS_CheckParm("help")) {
	    usage ();
	    do_png = 0;
	}

	if ((i = MS_CheckParm("which"))) {
		map = _argv[i+1][0] - '0';
	}

	if (MS_CheckParm("head")) {
	    show_head = 1;
	    do_png = 0;
	}

	if (MS_CheckParm("text")) {
	    do_text = 1;
	    do_png = 0;
	}

    read_maphead ();
    read_mapdata ();

    if (map > lastmap) {
	printf ("map %d not available, try a number in the range 0 to %d\n",
		map, lastmap);
	exit (-1);
    }

    if (map >= 0) iters = 1;
    else {  /* map == -1 */
	    map = 0;
	    iters = lastmap+1;
    }

    for (i = 0; i < iters; i++) {

	read_map (map);

	if (do_png) {
#ifdef __unix__
	    char fnamebuf[25];
	    snprintf (outfile, MAX_PATH, "%s/%s/maps",getenv("HOME"),CONFIG_USER_DIR);
	    CheckMakeDir(outfile);
	    sprintf (fnamebuf, "/map_%s_%d.png", extension, map);
	    strcat (outfile,fnamebuf);
#else
	    sprintf (outfile, "map_%s_%d.png", extension, map);
#endif
	    printf ("writing %s\n", outfile);

	    setup_png_file (outfile);
	    allocate_image ();
	    draw_walls ();
	    draw_grid ();
	    write_png_image ();
	}

	if (do_text) {
	    print_walls ();
	}

	map++;

    }
}

